/*  basic.h -- general macro definitions for all programs  */

#ifndef BASIC_H
#define BASIC_H

#if (defined(WIN32))
#  include<sys/types.h>
#  include<sys/timeb.h>
#  include <fcntl.h>
#  include <io.h>
#  define SET_BINARY_MODE(f) setmode(fileno(f), O_BINARY)
#else
#  include <sys/time.h>
#  include <sys/resource.h>
#  include <unistd.h>
#  define SET_BINARY_MODE(f)
#endif


#ifdef NO_ZLIB
#  include <stdio.h>
#  define FOPEN(path,mode) fopen(path,mode)
#  define FDOPEN(filedes,mode) fdopen(filedes,mode)
#  define FCLOSE(file)  fclose(file)
#  define FREAD(ptr,size,nmemb,file) fread(ptr,size,nmemb,file)
#  define FGETS(s,size,file) fgets(s,size,file)
#  define GETC(file) getc(file)
#  define UNGETC(c,file) ungetc(c,file)
typedef FILE *FILETYPE;
#else
#  include <zlib.h>
#  define FOPEN(path,mode) myfopen(path,mode)
#  define FDOPEN(filedes,mode) gzdopen(filedes,mode)
#  define FCLOSE(file)  gzclose(file)
#  define FREAD(ptr,size,nmemb,file) gzread(file,ptr,(size)*(nmemb))
#  define FGETS(s,size,file) gzgets(file,s,size)
#  define GETC(file) mygetc(file)
#  define UNGETC(c,file) myungetc(c,file)
typedef gzFile FILETYPE;
int mygetc( gzFile g );
int myungetc( int c, gzFile g );
FILETYPE myfopen( char *path, char *mode );
#endif

#include <limits.h>

/*
 * =========== Common Constants =============
 */
#ifndef PI
#define PI 3.1415926535897932385
#endif

#define ARCSEC_PER_RAD ((360 * 60 * 60) / (2.0 * PI))
#define METERS_PER_ARCSEC (1.0e7 / (90 * 60 * 60))
#define METERS_PER_FEET 0.3048

/*
 * =========== Common Macros =============
 */
#ifndef MAX
#define MAX(a, b) (((a)>(b))?(a):(b))
#endif

#ifndef MIN
#define MIN(a, b) (((a)>(b))?(b):(a))
#endif

#ifndef ABS
#define ABS(x) (((x)<0)?(-(x)):(x))
#endif

#ifndef ROUND
#define ROUND(a) ((int)((a)+0.5))
#endif

#define MIDPOINT(p1,p2,m) \
  (m)[0] = ((p1)[0] + (p2)[0]) / 2; \
  (m)[1] = ((p1)[1] + (p2)[1]) / 2;

#define LEFT_TURN(a, b, c) \
(((b)[1]-(a)[1])*((c)[0]-(a)[0])-((b)[0]-(a)[0])*((c)[1]-(a)[1]) < 0)

#define LEFT_OR_NO_TURN(a, b, c) \
(((b)[1]-(a)[1])*((c)[0]-(a)[0])-((b)[0]-(a)[0])*((c)[1]-(a)[1]) <= 0)

/*
 * ============ Common Structures ============
 */

typedef struct { double u, v; } UV;
typedef int Coord2d[2];
typedef float Coord2f[2];
typedef float Coord3f[3];

typedef struct {
    double ua, ub, uc, ud;
    double va, vb, vc, vd;
} Transform;

/*
 * ============ Common prototypes ============
 */

/* trans.c */
Transform *createTransform(UV swS, UV nwS, UV neS, UV seS,
			   UV swT, UV nwT, UV neT, UV seT);
void transform(Transform *f, UV s, UV *t);
void utm2lola(int zone, UV utm, UV *lola);
void lola2utm(UV lola, UV *utm);
void datum2datum(int datumIn, UV lolaIn, int datumOut, UV *lolaOut);

/* utils.c */
void nospace(void);
double getTime(void);
FILETYPE open4read( char *path );

#endif   /* BASIC_H */
