/*  dem.h -- structures and prototype for reading DEM files  */

#ifndef DEM_H
#define DEM_H

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <limits.h>
#include <ctype.h>
#include "basic.h"

/*
 * ============== USGS DEM format ==============
 */
#define MAXROWS 1420		/* the biggest DEM size I've seen */
#define MAXCOLS 1201

#define NAD27 1
#define WGS72 2
#define WGS84 3
#define NAD83 4
#define OLD_HAWAII_DATUM 5
#define PUERTO_RICO_DATUM 6

#define GEO_SYSTEM 0
#define UTM_SYSTEM 1

typedef struct {
  int rows, cols;
  double dz2m;		/* scale from data units to meters */
  int dzoffset;		/* add this to data item then scale */
  unsigned short dzmax;	/* data item maximum (always >= 0) */

  /* information from "A" record */
  char name[145];
  int level;
  int elevation_pattern;
  int reference_system;
  int zone;
  double map_projection_parameter[15];	/* always zero */
  int ground_unit;
  int elevation_unit;
  int n_polygon;
  UV corner[4];
  double minElevation, maxElevation;
  double axis_angle;
  int accuracy;
  double resolution[3];
  double cellx, celly;
  int datum;		/* one of the above datums (default=NAD27) */

  /* information from "B" records */
  int rowsInProfile[MAXCOLS], colsInProfile[MAXCOLS];
  UV firstUVInProfile[MAXCOLS];
  double firstElevationInProfile[MAXCOLS];
  double minElevationInProfile[MAXCOLS], maxElevationInProfile[MAXCOLS];
  unsigned short data[MAXCOLS][MAXROWS];

  /* information from "C" record */
  int C2;
  int C2x, C2y, C2z;
  int C2sample_size;
  int C5;
  int C5x, C5y, C5z;
  int C5sample_size;
} Dem;

/*
 * ============== Elevation Model =================
 */
typedef struct {
  int width, height;
  double dz2m;
  int dzoffset;
  unsigned short dzmax;

  int zone;
  int datum;
  int ref_system;	/* either GEO_SYSTEM or UTM_SYSTEM */
  UV groundO;		/* minimum UTM or arcsec corner */
  UV groundDelta;	/* change in UTM or arcsec per change in x and y */

  /***** longtitude and latitude are in arc seconds****/
  /***** order is longtitude=U, latitude=V         ****/
  UV swUV, nwUV, neUV, seUV;	/* longitude,latitude of 4 corners */
  UV swXY, nwXY, neXY, seXY;	/* x,y of 4 corners */
  unsigned short *data;

  /* modified by Xiangyu Zhang */
  float *normal;
} Em;

/* modified by Xiangyu Zhang */
#define EMNORMAL(em,i,j,k) ((em)->normal[((int)(j)*(em)->width + (int)(i))*3+k])
#define EMDATA(em, i, j) ((em)->data[(int)(j)*(em)->width + (int)(i)])


/* ========== prototypes ================ */

/* dem.c */
Em *readEM(char *filename);
Dem *readDEM(char *filename);
void transEMlola2xy(Em *em, UV lola, UV *xy);
void transEMxy2lola(Em *em, UV xy, UV *lola);



#endif	/* DEM_H */
