/*  infov.c -- field-of-view (FOV) calculations  */

#include "topo.h"

static int inFOVpt(Coord2d p)
{
  if( fovType == WEDGE ) {
    return( fovLnorm[0]*(p[0]-eye[0])+fovLnorm[1]*(p[1]-eye[1]) <= 0 &&
	    fovRnorm[0]*(p[0]-eye[0])+fovRnorm[1]*(p[1]-eye[1]) <= 0 );
  } else if( fovType == FOUR_GON ) {
    return( LEFT_TURN(p, fovLLpt, fovLRpt) &&
	    LEFT_TURN(p, fovLRpt, fovURpt) &&
	    LEFT_TURN(p, fovURpt, fovULpt) &&
	    LEFT_TURN(p, fovULpt, fovLLpt) );
  } else {
    return( LEFT_TURN(p, fovLLpt, eye) &&
	    LEFT_TURN(p, eye, fovLRpt) &&
	    LEFT_TURN(p, fovLRpt, fovURpt) &&
	    LEFT_TURN(p, fovURpt, fovULpt) &&
	    LEFT_TURN(p, fovULpt, fovLLpt) );
  }
}

static int crossFOV(Coord2d p, Coord2d q)
{
  double tin=0.0, tout=1.0;
  double num,den;

#define CLIP(N,R)\
  num = N[0]*(R[0]-p[0])+N[1]*(R[1]-p[1]);\
  den = N[0]*(q[0]-p[0])+N[1]*(q[1]-p[1]);\
  if( den < 0 ) {\
    tin = MAX(tin, num/den);\
  } else if( den > 0 ) {\
    tout = MIN(tout, num/den);\
  } else if( num < 0 )\
    return 0;\
  if( tin > tout ) return 0;

  CLIP(fovLnorm, fovLLpt);
  CLIP(fovRnorm, fovLRpt);
  if( fovType == WEDGE ) return 1;
  CLIP(fovTnorm, fovULpt);
  CLIP(fovBLnorm, fovLLpt);
  if( fovType == FOUR_GON ) return 1;
  CLIP(fovBRnorm, fovLRpt);
  return 1;
}

int inFOV(Coord2d p1, Coord2d p2, Coord2d p3)
{
  int a, b, c;

  a = inFOVpt(p1);
  b = inFOVpt(p2);
  c = inFOVpt(p3);
  if( a && b && c ) return ALLINFOV;
  if( a || b || c ) return PARTINFOV;

  /* All tri points are outside FOV.  Does tri cover part of FOV? */ 
  if( crossFOV(p1,p2) ) return PARTINFOV;
  if( crossFOV(p2,p3) ) return PARTINFOV;
  if( crossFOV(p3,p1) ) return PARTINFOV;

  /* Perhaps all of FOV is in triangle (check eye position) */
  if( LEFT_TURN(eye, p1, p2) &&
      LEFT_TURN(eye, p2, p3) &&
      LEFT_TURN(eye, p3, p1) )
    return PARTINFOV;
  return 0;
}

/*  triInFOV(t) -- is triangle t in field of view?  */
int triInFOV(TriIndex t)
{
  Coord2d p1, p2, p3;

  getVerts(t, p1, p2, p3);
  return inFOV(p1, p2, p3);
}



/*  inROD(p1, p2, p3)  */

int inROD(Coord2d p1, Coord2d p2, Coord2d p3)
{
  int r;
  int txmax = MAX(p1[0], MAX(p2[0], p3[0]));
  int txmin = MIN(p1[0], MIN(p2[0], p3[0]));
  int tymax = MAX(p1[1], MAX(p2[1], p3[1]));
  int tymin = MIN(p1[1], MIN(p2[1], p3[1]));

  if (txmax > gemXmax - 1) {
    if (txmin >= gemXmax - 1) {
      r = 0;
    } else {
      r = PARTINROD;
    }
  } else if (tymax > gemYmax - 1) {
    if (tymin >= gemYmax - 1) {
      r = 0;
    } else {
      r = PARTINROD;
    }
  } else {
    r = ALLINROD;
  }

  return r;
}
