/*  tri.c -- initialize triangle hierarchy  */

#include "topo.h"

static double x_per_y, y_per_x, errPerEm;



static int calcErr(Coord2d p1, Coord2d p2, Coord2d p3)
{
  int x1, x2, x3, y1, y2, y3;
  double z1, z2, z3;
  double ystart, yend, xstart, xend;
  double deltaxstart, deltaxend, deltaystart, deltayend;
  int xmini, xmaxi, ymini, ymaxi;
  double z, zstart, zend, deltazstart, deltazend, deltaz;
  double max_eps, eps;
  int x, y;
  int xstarti, xendi, ystarti, yendi;
	
  x1 = p1[0];
  y1 = p1[1];
  z1 = EMDATA(gem, x1, y1);
  x2 = p2[0];
  y2 = p2[1];
  z2 = EMDATA(gem, x2, y2);
  x3 = p3[0];
  y3 = p3[1];
  z3 = EMDATA(gem, x3, y3);
	
  assert(x1 < gemXmax);
  assert(x2 < gemXmax);
  assert(x3 < gemXmax);
  assert(y1 < gemYmax);
  assert(y2 < gemYmax);
  assert(y3 < gemYmax);
	
  max_eps = 0.0;
	
  if (y1 == y2) {
    if (x1 > x2) {
      xmini = x2;
      xmaxi = x1;
      ymini = y3;
      ymaxi = y1;
			
      xstart = x3;
      xend = xstart;
      deltaxstart = -x_per_y;
      deltaxend = x_per_y;
      zstart = z3;
      zend = z3;
      deltazstart = (z2 - zstart) / (double)(y1 - y3);
      deltazend = (z1 - zend) / (double)(y1 - y3);
    } else {
      xmini = x1;
      xmaxi = x2;
      ymini = y1;
      ymaxi = y3;
			
      xstart = x1;
      xend = x2;
      deltaxstart = x_per_y;
      deltaxend = -x_per_y;
      zstart = z1;
      zend = z2;
      deltazstart = (z3 - zstart) / (double)(y3 - y1);
      deltazend = (z3 - zend) / (double)(y3 - y1);
    }
		
    for (y = ymini; y <= ymaxi; ++y) {
      xstarti = MAX(xmini, ROUND(xstart));
      xendi = MIN(xmaxi, ROUND(xend));
      z = zstart;
      if (xstarti != xendi) {
	deltaz = (zend - zstart) / (xendi - xstarti);
      } else {
	deltaz = 0.0;
      }
      for (x = xstarti; x <= xendi; ++x) {
	eps = ABS(EMDATA(gem, x, y) - z);
	if (eps > max_eps) {
	  max_eps = eps;
	}
	z = z + deltaz;
      }
      zstart = zstart + deltazstart;
      zend = zend + deltazend;
      xstart = xstart + deltaxstart;
      xend = xend + deltaxend;
    }
  } else {
		
    if (x1 == x2) {
      if (y1 > y2) {
	xmini = x1;
	xmaxi = x3;
	ymini = y2;
	ymaxi = y1;
				
	ystart = y2;
	yend = y1;
	deltaystart = y_per_x;
	deltayend = -y_per_x;
	zstart = z2;
	zend = z1;
	deltazstart = (z3 - zstart) / (double)(x3 - x1);
	deltazend = (z3 - zend) / (double)(x3 - x1);
      } else {
	xmini = x3;
	xmaxi = x1;
	ymini = y1;
	ymaxi = y2;
				
	ystart = y3;
	yend = ystart;
	deltaystart = -y_per_x;
	deltayend = y_per_x;
	zstart = z3;
	zend = z3;
	deltazstart = (z1 - zstart) / (double)(x1 - x3);
	deltazend = (z2 - zend) / (double)(x1 - x3);
      }
			
    } else {
      if (x1 < x2) {
	if (y1 > y2) {
	  xmini = x1;
	  xmaxi = x2;
	  ymini = y2;
	  ymaxi = y3;
					
	  ystart = y1;
	  yend = ystart;
	  deltaystart = -y_per_x;
	  deltayend = 0.0;
	  zstart = z1;
	  zend = z1;
	  deltazstart = (z2 - zstart) / (double)(x3 - x1);
	  deltazend = (z3 - zend) / (double)(x3 - x1);
	} else {
	  xmini = x1;
	  xmaxi = x2;
	  ymini = y1;
	  ymaxi = y3;
					
	  ystart = y1;
	  yend = y3;
	  deltaystart = y_per_x;
	  deltayend = 0.0;
	  zstart = z1;
	  zend = z3;
	  deltazstart = (z2 - zstart) / (double)(x2 - x3);
	  deltazend = (z2 - zend) / (double)(x2 - x3);
	}
      } else {
	if (y1 > y2) {
	  xmini = x2;
	  xmaxi = x3;
	  ymini = y3;
	  ymaxi = y1;
					
	  ystart = y2;
	  yend = ystart;
	  deltaystart = 0.0;
	  deltayend = y_per_x;
	  zstart = z2;
	  zend = z2;
	  deltazstart = (z3 - zstart) / (double)(x3 - x2);
	  deltazend = (z1 - zend) / (double)(x3 - x2);
	} else {
	  xmini = x3;
	  xmaxi = x1;
	  ymini = y3;
	  ymaxi = y2;
					
	  ystart = y3;
	  yend = y2;
	  deltaystart = 0.0;
	  deltayend = -y_per_x;
	  zstart = z3;
	  zend = z2;
	  deltazstart = (z1 - zstart) / (double)(x1 - x3);
	  deltazend = (z1 - zend) / (double)(x1 - x3);
	}
      }
    }
    for (x = xmini; x <= xmaxi; ++x) {
      ystarti = MAX(ymini, ROUND(ystart));
      yendi = MIN(ymaxi, ROUND(yend));
      z = zstart;
      if (ystarti != yendi) {
	deltaz = (zend - zstart) / (yendi - ystarti);
      } else {
	deltaz = 0.0;
      }
      for (y = ystarti; y <= yendi; ++y) {
	eps = ABS(EMDATA(gem, x, y) - z);
	if (eps > max_eps) {
	  max_eps = eps;
	}
	z = z + deltaz;
      }
      zstart = zstart + deltazstart;
      zend = zend + deltazend;
      ystart = ystart + deltaystart;
      yend = yend + deltayend;
    }
  }
	
  return ROUND(max_eps * errPerEm);
}

/* setCoord(p, q) -- copy Coord2d q into Coord2d p */
void setCoord(Coord2d p, Coord2d q)
{
    p[0] = q[0];
    p[1] = q[1];
}

static void getV(int p, Coord2d p1, Coord2d p2, Coord2d p3)
/*
 * Note: p is TriIndex + 1 so that the bit sequence describes a path in the
 *       tree.
 *       Don't use getV.  Use getVerts.
 */
{
    switch(p) {
    case 0:
    case 1:
    case 2:
	setCoord(p1, sw);
	setCoord(p2, ne);
	setCoord(p3, nw);
	return;
    case 3:
	setCoord(p1, ne);
	setCoord(p2, sw);
	setCoord(p3, se);
	return;
    default:
	if (p & 1) {
	    getV(p>>1, p3, p1, p2);
	    MIDPOINT(p1, p3, p3);
	    return;
	} else {
	    getV(p>>1, p2, p3, p1);
	    MIDPOINT(p2, p3, p3);
	    return;
	}
    }
}



void getVerts(TriIndex t, Coord2d p1, Coord2d p2, Coord2d p3)
{
    getV(t + 1, p1, p2, p3);
}


static int ceilingLog2(int x)
{
  int m = 1;
  int i = 0;
	
  while (m < x && m < INT_MAX) {
    m <<= 1;
    ++i;
  }
  return i;
}

static int next2Power(int x)
{
  int m = 1;
  while (m < x && m < INT_MAX)
    m <<= 1;
  return m;
}

static void makeDFS(Hier *hp, TriIndex t, Coord2d p1, Coord2d p2, Coord2d p3)
{
  Coord2d mid;
  int r;
	
  r = inROD(p1, p2, p3);
  hp->info[t] = r;
  if (ISLEAF(hp,t)) {
    hp->err[t] = 0;
  } else {
    if (r == 0) {
      hp->err[t] = 0;
    } else if (r == PARTINROD) {
      hp->err[t] = MAXERR;
    } else {
      hp->err[t] = calcErr(p1, p2, p3);
    }
    MIDPOINT(p1, p2, mid);
    makeDFS(hp, LCHILD(t), p3, p1, mid);
    makeDFS(hp, RCHILD(t), p2, p3, mid);
  }
}



/*  initHier() -- initialize triangle hierarchy  */

Hier *initHier(void)
{
  Hier *hp;
  int dim;
	
  x_per_y = (double)gemXmax / (double)gemYmax;
  y_per_x = (double)gemYmax / (double)gemXmax;
	
  assert(gem->dzmax > 0);
  errPerEm = (double)MAXERR / (double)gem->dzmax;

  hp = (Hier *)malloc(sizeof(Hier));
  if (hp == NULL) {
    fprintf(stderr, "Can't allocate space for hierarchy.\n");
    exit(1);
  }
  dim = next2Power(MAX(gemXmax, gemYmax) - 1) + 1;
  hp->size = 4*(dim-1)*(dim-1)-1;
  hp->depth = 2*ceilingLog2(dim-1) + 1;
  hp->numSurfTri = 0;
  hp->err = (unsigned short *)malloc(hp->size * sizeof(short));
  hp->info = (unsigned char *)malloc(hp->size * sizeof(char));
  if (hp->err == NULL || hp->info == NULL) {
    fprintf(stderr, "Can't allocate space for %d nodes.\n", hp->size);
    exit(1);
  }
	
  fprintf(stderr, "|Hierarchy|+|Elevation Model|=%d bytes.\n",
	  (int)(gem->width * gem->height * sizeof(*(gem->data))+
		hp->size*(sizeof(*(hp->err))+sizeof(*(hp->info)))));

  sw[0] = 0;
  sw[1] = 0;
	
  se[0] = dim - 1;
  se[1] = 0;
	
  ne[0] = dim - 1;
  ne[1] = dim - 1;
	
  nw[0] = 0;
  nw[1] = dim - 1;
	
  makeDFS(hp, LCHILD(0), sw, ne, nw);	/* NW corner of square */
  makeDFS(hp, RCHILD(0), ne, sw, se);	/* SE corner of square */
  hp->err[0] = MAXERR;
  hp->info[0] = 0;
	
  return hp;
}
